@@ -260,7 +260,7 @@ src_present_real_sro_for_speech (SRObject *obj, gchar *reason)
 		strcmp (reason, "object:context-switched") == 0)
 	    {
 		if ( role == SR_ROLE_STATUS_BAR && strcmp (reason , "focus:") == 0)
-		    src_speech_send_chunk (txt, SRC_SPEECH_PRIORITY_SYSTEM, FALSE);
+		    src_speech_send_chunk (txt, SRC_SPEECH_PRIORITY_SYSTEM, TRUE);
 		else if ((role == SR_ROLE_TABLE || role == SR_ROLE_TREE_TABLE) &&
 			strcmp (reason , "focus:") == 0)
 		    src_speech_send_chunk (txt, SRC_SPEECH_PRIORITY_IDLE, FALSE);
