@@ -25,6 +25,7 @@
 #include "cspi/spi.h"
 #include "SRObject.h"
 #include "string.h"
+#include  "gdk/gdk.h"
 
 #define	SRL_MAX_CLIENTS			1
 #define	SRL_LOG
@@ -68,6 +69,12 @@
 #define SRL_EVENT_NAME_UNKNOWN			"unknown"
 #define SRL_EVENT_NAME_ACTIVE_DESCENDANT_CHANGED "object:active-descendant-changed"
 #define SRL_EVENT_NAME_NAME_CHANGED		"object:property-change:accessible-name"
+#define SRL_EVENT_NAME_VISIBLE_DATA_CHANGED	"object:visible-data-changed"
+#define SRL_EVENT_NAME_CHILDREN_ADDED           "object:children-changed:add"
+#define SRL_EVENT_NAME_CHILDREN_REMOVED         "object:children-changed:remove"
+#define SRL_EVENT_NAME_TAB_ADDED                "object:tab-added"
+#define SRL_EVENT_NAME_TAB_REMOVED              "object:tab-removed"
+#define SRL_EVENT_NAME_CONTEXT_SWITCHED         "object:context-switched"
 
 #define SRL_TOOLKIT_MOZILLA			"mozilla"
 #define SRL_APP_METACITY			"metacity"
@@ -88,7 +95,7 @@ typedef enum _SRLEventType
     SRL_EVENT_TEXT_SELECTION_CHANGED,
     SRL_EVENT_VALUE_CHANGED,
     SRL_EVENT_STATE_CHECKED,
-    SRL_EVENT_STATE_SELECTED, /*FIXME: really needed???? */
+    SRL_EVENT_STATE_SELECTED, 
     SRL_EVENT_STATE_EXPANDED,
     SRL_EVENT_STATE_SHOWING,
     SRL_EVENT_SELECTION_CHANGED,
@@ -105,6 +112,12 @@ typedef enum _SRLEventType
     SRL_EVENT_TOOLTIP_HIDE,
     SRL_EVENT_ACTIVE_DESCENDANT_CHANGED,
     SRL_EVENT_NAME_CHANGED,
+    SRL_EVENT_VISIBLE_DATA_CHANGED,
+    SRL_EVENT_CHILDREN_ADDED,
+    SRL_EVENT_CHILDREN_REMOVED,
+    SRL_EVENT_TAB_ADDED,
+    SRL_EVENT_TAB_REMOVED,
+    SRL_EVENT_CONTEXT_SWITCHED,
     SRL_EVENT_WINDOW_RENAME
 }SRLEventType;
 
@@ -113,6 +126,7 @@ typedef enum _SRLEventPriority
     SRL_EVENT_PRIORITY_MOUSE,
     SRL_EVENT_PRIORITY_TOOLTIP,
     SRL_EVENT_PRIORITY_FOCUS,
+    SRL_EVENT_PRIORITY_WATCHED,
     SRL_EVENT_PRIORITY_WINDOW,
     SRL_EVENT_PRIORITY_LAST,
     SRL_EVENT_PRIORITY_UNKNOWN = SRL_EVENT_PRIORITY_LAST
@@ -147,7 +161,10 @@ static SRLEventTypeEventName srl_events_type_name[] =
 	{SRL_EVENT_WINDOW_DEACTIVATE,		SRL_EVENT_NAME_WINDOW_DEACTIVATE},
 	{SRL_EVENT_MOUSE_ABS,			SRL_EVENT_NAME_MOUSE_ABS},
 	{SRL_EVENT_ACTIVE_DESCENDANT_CHANGED, 	SRL_EVENT_NAME_ACTIVE_DESCENDANT_CHANGED},
-	{SRL_EVENT_NAME_CHANGED,		SRL_EVENT_NAME_NAME_CHANGED}
+	{SRL_EVENT_NAME_CHANGED,		SRL_EVENT_NAME_NAME_CHANGED},
+	{SRL_EVENT_VISIBLE_DATA_CHANGED,	SRL_EVENT_NAME_VISIBLE_DATA_CHANGED},
+	{SRL_EVENT_CHILDREN_ADDED,              SRL_EVENT_NAME_CHILDREN_ADDED},
+	{SRL_EVENT_CHILDREN_REMOVED,            SRL_EVENT_NAME_CHILDREN_REMOVED}
     };
 
 typedef struct _SRLEvent
@@ -176,8 +193,11 @@ static gboolean			srl_idle_installed	= FALSE;
 static GQueue			*srl_event_queue	= NULL;
 static AccessibleEventListener 	*srl_event_listeners[G_N_ELEMENTS (srl_events_type_name)];
 static Accessible		*srl_last_focus		= NULL;
+static Accessible		*srl_last_focus2	= NULL;
 static Accessible		*srl_watched_acc	= NULL;
 static Accessible 		*srl_last_table 	= NULL;
+static Accessible 		*srl_last_context 	= NULL;
+static Accessible		*srl_last_create	= NULL;
 /*FIXME:next varaible should be in SRObject.c file */
 Accessible 			*srl_last_edit		= NULL;
 
@@ -493,6 +513,7 @@ srle_free (SRLEvent *event)
 
     if (event->acc)
 	Accessible_unref (event->acc);
+    if (event->acc_ev)
     AccessibleEvent_unref (event->acc_ev);
     g_free (event);
 }
@@ -599,8 +620,11 @@ srle_set_acc (SRLEvent *event)
 		Accessible_unref (parent);
 	}
     }
-    else if (srle_has_type (event, SRL_EVENT_FOCUS) &&
-	srle_acc_has_state (event, SPI_STATE_MANAGES_DESCENDANTS))
+    else if ((srle_has_type (event, SRL_EVENT_FOCUS) &&
+		    srle_acc_has_state (event, SPI_STATE_MANAGES_DESCENDANTS)) ||
+	     ( (srle_has_type (event, SRL_EVENT_SELECTION_CHANGED)  ||
+	           srle_has_type (event, SRL_EVENT_FOCUS)) &&
+		srle_acc_has_role (event, SPI_ROLE_LAYERED_PANE)))
     {
 	AccessibleSelection *selection;
 	selection = Accessible_getSelection ((Accessible*)srle_get_acc (event));
@@ -630,6 +654,11 @@ srle_set_acc (SRLEvent *event)
     {
 	event->acc = AccessibleActiveDescendantChangedEvent_getActiveDescendant (event->acc_ev);
     }
+    if (srle_has_type (event, SRL_EVENT_CHILDREN_ADDED) ||
+        srle_has_type (event, SRL_EVENT_CHILDREN_REMOVED))
+    {
+	event->acc = AccessibleChildChangedEvent_getChildAccessible (event->acc_ev);
+    }	
     if (srle_has_type (event, SRL_EVENT_LINK_SELECTED))
     {
 	if (srl_acc_has_toolkit (srle_get_acc (event), SRL_TOOLKIT_MOZILLA))
@@ -660,11 +689,16 @@ srle_get_reason (SRLEvent *event)
 
     srl_assert (event);
 
+    if (srle_has_type (event, SRL_EVENT_TAB_ADDED))
+	return SRL_EVENT_NAME_TAB_ADDED;
+    if (srle_has_type (event, SRL_EVENT_TAB_REMOVED))
+	return SRL_EVENT_NAME_TAB_REMOVED;		
+	
     if (srle_has_type (event, SRL_EVENT_SELECTION_CHANGED) &&
 	srle_acc_has_role (event, SPI_ROLE_COMBO_BOX))
 	return SRL_EVENT_NAME_CONTENT_CHANGED;
     if (srle_has_type (event, SRL_EVENT_WINDOW_ACTIVATE))
-	return SRL_EVENT_NAME_WINDOW_SWITCH;
+	return SRL_EVENT_NAME_WINDOW_SWITCH;		
 
     for (i = 0; i < G_N_ELEMENTS (srl_events_type_name); i++)
     	if (srle_has_type (event, srl_events_type_name[i].type))
@@ -684,6 +718,8 @@ srle_get_reason (SRLEvent *event)
 	return SRL_EVENT_NAME_FOCUS;
     if (srle_has_type (event, SRL_EVENT_FOCUS2))
 	return SRL_EVENT_NAME_FOCUS;
+    if (srle_has_type (event, SRL_EVENT_CONTEXT_SWITCHED))
+	return SRL_EVENT_NAME_CONTEXT_SWITCHED;
 
     srl_assert_not_reached ();
     return SRL_EVENT_NAME_UNKNOWN;
@@ -694,7 +730,8 @@ srl_is_window_event (SRLEvent *event)
 {
     srl_assert (event);
     
-    if (srle_acc_has_role (event, SPI_ROLE_TOOL_TIP))
+    if (srle_acc_has_role (event, SPI_ROLE_TOOL_TIP) ||
+	srle_acc_has_role (event, SPI_ROLE_WINDOW))
 	return FALSE;
     return  event->type == SRL_EVENT_WINDOW_CREATE   ||
 	    event->type == SRL_EVENT_WINDOW_DESTROY  ||
@@ -704,15 +741,34 @@ srl_is_window_event (SRLEvent *event)
 	    event->type == SRL_EVENT_WINDOW_ACTIVATE ||
 	    event->type == SRL_EVENT_WINDOW_TITLELIZE||
 	    event->type == SRL_EVENT_WINDOW_DEACTIVATE ||
-	    event->type == SRL_EVENT_WINDOW_RENAME;
+	    event->type == SRL_EVENT_WINDOW_RENAME ||
+	    event->type == SRL_EVENT_TAB_ADDED ||
+	    event->type == SRL_EVENT_CONTEXT_SWITCHED ||
+	    event->type == SRL_EVENT_TAB_REMOVED;
+}
+static gboolean
+srle_is_for_watched_acc (SRLEvent *event)
+{
+    srl_assert (event);
+    
+    return srl_watched_acc == event->acc_ev->source;
 }
 
+static gboolean
+srle_is_for_focused_acc (SRLEvent *event)
+{
+    srl_assert (event);
+
+    return srl_last_focus == srle_get_acc (event);
+}
+
+/*
 static gboolean
 srl_is_focus_event (SRLEvent *event)
 {
     srl_assert (event);
     
-    return  event->type == SRL_EVENT_FOCUS		||
+    return  (event->type == SRL_EVENT_FOCUS		||
 	    event->type == SRL_EVENT_FOCUS1		||
 	    event->type == SRL_EVENT_FOCUS2		||
 	    event->type == SRL_EVENT_LINK_SELECTED	||
@@ -726,7 +782,34 @@ srl_is_focus_event (SRLEvent *event)
 	    event->type == SRL_EVENT_STATE_SELECTED	||
 	    event->type == SRL_EVENT_STATE_EXPANDED	||
 	    event->type == SRL_EVENT_STATE_SHOWING	||
-	    event->type == SRL_EVENT_SELECTION_CHANGED;
+	    event->type == SRL_EVENT_SELECTION_CHANGED	||
+	    event->type == SRL_EVENT_VISIBLE_DATA_CHANGED) &&
+	    srle_is_for_focused_acc (event);
+}
+*/
+
+static gboolean
+srl_is_watched_event (SRLEvent *event)
+{
+    srl_assert (event);
+    
+    return  (event->type == SRL_EVENT_FOCUS		||
+	    event->type == SRL_EVENT_FOCUS1		||
+	    event->type == SRL_EVENT_FOCUS2		||
+	    event->type == SRL_EVENT_LINK_SELECTED	||
+	    event->type == SRL_EVENT_TEXT_CARET_MOVED	||
+	    event->type == SRL_EVENT_TEXT_CHANGED_INSERT||
+	    event->type == SRL_EVENT_TEXT_CHANGED_DELETE||
+	    event->type == SRL_EVENT_TEXT_SELECTION_CHANGED||
+	    event->type == SRL_EVENT_VALUE_CHANGED	||
+	    event->type == SRL_EVENT_NAME_CHANGED	||
+	    event->type == SRL_EVENT_STATE_CHECKED	||
+	    event->type == SRL_EVENT_STATE_SELECTED	||
+	    event->type == SRL_EVENT_STATE_EXPANDED	||
+	    event->type == SRL_EVENT_STATE_SHOWING	||
+	    event->type == SRL_EVENT_SELECTION_CHANGED	||
+	    event->type == SRL_EVENT_VISIBLE_DATA_CHANGED) &&
+	    srle_is_for_watched_acc (event);
 }
 
 static gboolean
@@ -749,17 +832,16 @@ static SRLEventPriority
 srle_get_priority (SRLEvent *event)
 {
     srl_assert (event);
-    
-    if (srl_is_focus_event (event))
-	return SRL_EVENT_PRIORITY_FOCUS;
+	
+    if (srl_is_watched_event (event))
+	return SRL_EVENT_PRIORITY_WATCHED;
     if (srl_is_window_event (event))
 	return SRL_EVENT_PRIORITY_WINDOW;
     if (srl_is_mouse_event (event))
 	return SRL_EVENT_PRIORITY_MOUSE;
     if (srl_is_tooltip_event (event))
 	return SRL_EVENT_PRIORITY_TOOLTIP;
-    srl_assert_not_reached ();
-    return SRL_EVENT_PRIORITY_UNKNOWN;
+    return SRL_EVENT_PRIORITY_FOCUS;
 }
 
 static gboolean
@@ -923,7 +1005,7 @@ srl_notify_clients_obj (SRLEvent *event,
 		sro_set_difference (obj, diff);
 	    SPI_freeString (diff);
 	}
-	if (srl_is_window_event (event))
+	if (srl_is_window_event (event) && event->type != SRL_EVENT_CONTEXT_SWITCHED)
 	{
 	    gchar *name = AccessibleWindowEvent_getTitleString (event->acc_ev);
 	    if (name)
@@ -1018,6 +1100,8 @@ srl_report_event_to_clients (gpointer data)
 {
     srl_report_obj_event_to_clients (SRL_EVENT_PRIORITY_WINDOW, 
 					    SR_EVENT_WINDOW);    
+    srl_report_obj_event_to_clients (SRL_EVENT_PRIORITY_WATCHED,
+					    SR_EVENT_SRO);
     srl_report_obj_event_to_clients (SRL_EVENT_PRIORITY_FOCUS,
 					    SR_EVENT_SRO);
     srl_report_obj_event_to_clients (SRL_EVENT_PRIORITY_TOOLTIP,
@@ -1118,15 +1202,31 @@ srl_acc_has_toolkit (const Accessible *acc,
     return rv;    
 }
 
-
 static gboolean
-srle_is_for_focused_acc (SRLEvent *event)
+srle_is_for_focused_ancestor_acc (SRLEvent *event)
 {
+    Accessible *acc;
+    gboolean rv = FALSE;
     srl_assert (event);
 
-    return srl_last_focus == srle_get_acc (event);
+    acc = event->acc_ev->source;
+    Accessible_ref (acc);
+    while (acc)
+    {
+	Accessible *tmp;
+	rv = srl_acc_has_state (acc, SPI_STATE_FOCUSED);
+	if (rv)
+	    break;
+	tmp = acc;
+	acc = Accessible_getParent (tmp);
+	Accessible_unref (tmp);
+    }
+    if (acc)
+	Accessible_unref (acc);
+    return rv;
 }
 
+
 static gboolean
 srl_event_is_for_metacity (SRLEvent *event)
 {
@@ -1155,7 +1255,6 @@ srl_event_is_for_metacity (SRLEvent *event)
     return rv;
 }
 
-
 static gboolean
 srle_change_type (SRLEvent *event)
 {
@@ -1184,9 +1283,11 @@ srle_change_type (SRLEvent *event)
 	    srl_acc_has_toolkit (srle_get_acc (event), SRL_TOOLKIT_MOZILLA))
 	    event->type = SRL_EVENT_FOCUS;
 */
-	if (!srle_is_for_focused_acc (event))
+	if (!srle_is_for_focused_acc (event) &&
+	    !srle_is_for_watched_acc (event) &&
+	    srle_is_for_focused_ancestor_acc (event))
 	    event->type = SRL_EVENT_FOCUS2;
-	if (last_link_index != new_link_index)
+	if (last_link_index != new_link_index && new_link_index != -1)
 	    event->type = SRL_EVENT_FOCUS;
 	last_link_index = new_link_index;
     }
@@ -1211,6 +1312,26 @@ srle_change_type (SRLEvent *event)
 	if (srl_acc_has_toolkit (srle_get_acc (event), SRL_TOOLKIT_MOZILLA))
 	    event->type = SRL_EVENT_FOCUS;
     }
+    else if (srle_has_type (event, SRL_EVENT_SELECTION_CHANGED))
+    {
+	if ((Accessible_getRole (event->acc_ev->source) == SPI_ROLE_LAYERED_PANE) &&
+	          (Accessible_getRole (event->acc) == SPI_ROLE_ICON))
+	    event->type = SRL_EVENT_FOCUS;
+    }
+    
+    if (srle_has_type (event, SRL_EVENT_CHILDREN_ADDED))
+	event->type = SRL_EVENT_TAB_ADDED;
+    else if (srle_has_type (event, SRL_EVENT_CHILDREN_REMOVED))
+	event->type = SRL_EVENT_TAB_REMOVED;
+	
+    /* Convert the "state-changed:selected" event for a _selected_ page-tab
+       into a "focus:" event */
+    if (srle_has_type (event, SRL_EVENT_STATE_SELECTED) &&
+        srle_acc_has_role (event, SPI_ROLE_PAGE_TAB)    &&
+	srl_acc_has_state (event->acc_ev->source, SPI_STATE_SELECTED))
+    {
+	event->type = SRL_EVENT_FOCUS;
+    }	       	
     
     return TRUE;
 }
@@ -1348,7 +1469,8 @@ srl_text_event_is_reported (SRLEvent *event,
     else
 	rv = TRUE;
 
-    if (srle_acc_has_role (event, SPI_ROLE_SPIN_BUTTON))
+    if (srle_acc_has_role (event, SPI_ROLE_SPIN_BUTTON) ||
+        srle_acc_has_role (event, SPI_ROLE_SLIDER))
         rv = !srl_info_is_value_changed (info, info2);
 
     /* when a text-changed event for child of combos is emitted,
@@ -1431,13 +1553,6 @@ srl_selection_event_is_reported (SRLEvent *event)
     return rv;
 }
 
-static gboolean
-srle_is_for_watched_acc (SRLEvent *event)
-{
-    srl_assert (event);
-    
-    return srl_watched_acc == event->acc_ev->source;
-}
 
 
 static gboolean
@@ -1450,12 +1565,37 @@ srl_event_is_reported (SRLEvent *event)
     srl_assert (event);
     last_info = srle_is_for_focused_acc (event) ? &last_info_focus : &last_info_watched;
 
-    if (srl_is_window_event (event) 	||
-	srl_is_mouse_event (event)	||
+    if (srl_is_window_event (event))
+    {
+	rv = TRUE;
+	if (srle_has_type (event, SRL_EVENT_WINDOW_CREATE))
+	{
+	    if (srl_last_create)
+		Accessible_unref (srl_last_create);
+	    srl_last_create = event->acc_ev->source;
+	    Accessible_ref (srl_last_create);
+	}
+	else
+	{
+	    if (srl_last_create == event->acc_ev->source)
+		rv = FALSE;
+	}
+	if (srle_has_type (event, SRL_EVENT_WINDOW_ACTIVATE))
+	{
+	    if (srl_last_create != event->acc_ev->source)
+	    {
+		if (srl_last_create)
+		    Accessible_unref (srl_last_create);
+		srl_last_create = NULL;
+	    }
+	}
+    }
+    else if (srl_is_mouse_event (event)	||
 	srl_is_tooltip_event (event))
 	rv = TRUE;
     else if (srle_has_type (event, SRL_EVENT_FOCUS))
-	rv = srle_acc_has_state (event, SPI_STATE_FOCUSED);
+	rv = srle_acc_has_state (event, SPI_STATE_FOCUSED) ||
+	     srle_acc_has_state (event, SPI_STATE_TRANSIENT);
     else if (srle_has_type (event, SRL_EVENT_FOCUS1))
 	rv = TRUE;
     else if (srle_has_type (event, SRL_EVENT_FOCUS2))
@@ -1486,29 +1626,92 @@ srl_event_is_reported (SRLEvent *event)
 	    rv = TRUE;
 	else if (srle_has_type (event, SRL_EVENT_VALUE_CHANGED))
 	    rv = srle_acc_has_role (event, SPI_ROLE_SPIN_BUTTON) ||
-		    srle_acc_has_role (event, SPI_ROLE_SLIDER);
+		 srle_acc_has_role (event, SPI_ROLE_SLIDER);
+	else if (srle_has_type (event, SRL_EVENT_VISIBLE_DATA_CHANGED))
+	    rv = TRUE;
     }
 
     if (srle_has_type (event, SRL_EVENT_WINDOW_DEACTIVATE))
 	rv = FALSE;
 
-    if (rv && (srle_is_for_focused_acc (event) ||
-	      srle_is_for_watched_acc (event)))
+    if (rv && (srle_has_type (event, SRL_EVENT_FOCUS) ||
+	      srle_is_for_watched_acc (event) ||
+	      srle_is_for_focused_acc (event)))
     {
 	srl_assert (last_info);
-	srl_set_info (event->acc_ev->source, last_info);
+	if (srle_has_type (event, SRL_EVENT_TEXT_CHANGED_DELETE))
+	{
+	    last_info->text_caret_offset  = event->acc_ev->detail1;
+	    last_info->text_selections_count  = 0;
+	    last_info->text_char_count -= event->acc_ev->detail2;
+	    last_info->text_crt_selection_length = 0;
+	}
+	else
+	    srl_set_info (event->acc_ev->source, last_info);
 
-	if (srle_is_for_focused_acc (event))
-	    if (last_info != &last_info_focus)
-		last_info_focus = *last_info;
 	if (srle_is_for_watched_acc (event))
+	{
 	    if (last_info != &last_info_watched)
 		last_info_watched = *last_info;
+	}else
+	{
+	    if (last_info != &last_info_focus)
+		last_info_focus = *last_info;
+	}
     }
     
     if (srle_acc_has_role (event, SPI_ROLE_EDITBAR))
 	rv = TRUE;
-
+    
+    /*made the traversal in order to receive
+     "object:children-change:add" and "object:children-change:remove"
+     events for SPI_ROLE_PAGE_TAB_LIST objects
+     */
+    if (srle_has_type (event, SRL_EVENT_FOCUS) && 
+        srle_acc_has_role (event, SPI_ROLE_TERMINAL))
+    {
+	gint childrencnt = 0;		
+        Accessible *parent = Accessible_getParent (event->acc_ev->source);
+	Accessible *child = NULL;
+	
+        while (Accessible_getRole (parent) != SPI_ROLE_PAGE_TAB_LIST)
+	    parent = Accessible_getParent (parent);
+	
+	if (parent)
+	{   
+	    gint i;
+	    childrencnt = Accessible_getChildCount (parent);
+	    for (i = 0; i < childrencnt; i++)
+	    {
+		child = Accessible_getChildAtIndex (parent, i);
+	    }
+	    
+	    if (child)
+		Accessible_unref (child);
+	    Accessible_unref (parent);	
+	}
+    }
+    if ((srle_has_type (event, SRL_EVENT_TAB_ADDED) || 
+         srle_has_type (event, SRL_EVENT_TAB_REMOVED))) 
+	rv = srle_acc_has_role (event, SPI_ROLE_PAGE_TAB);
+	
+    /*In some situations we have two 'focus:' events for a page-tab 
+     (an "object:state-changed:selected" event which was converted into
+      a focus event _and_ a "focus:" event) => report a single "focus:" event*/	
+    if (srle_has_type (event, SRL_EVENT_FOCUS) &&
+        srle_acc_has_role (event, SPI_ROLE_PAGE_TAB))
+    {
+	if (srl_last_focus2 == event->acc_ev->source)
+	    rv = FALSE;
+	else
+	    rv = TRUE;    
+    }	     
+    
+    if (srle_has_type (event, SRL_EVENT_TEXT_CARET_MOVED) &&
+        (event->acc_ev->detail1 == -1 ||
+	 event->acc_ev->detail2 == -1))
+	rv = FALSE;			 
+	
     return rv;
 }
 
@@ -1523,6 +1726,159 @@ srl_set_last_focus (Accessible *acc)
     return TRUE;
 }
 
+static gboolean
+srl_set_last_focus2 (Accessible *acc)
+{
+    if (srl_last_focus2)
+	Accessible_unref (srl_last_focus2);
+    srl_last_focus2 = acc;
+    if (srl_last_focus2)
+	Accessible_ref (srl_last_focus2);
+    return TRUE;
+}
+
+static gboolean
+srl_is_label_for (Accessible *label, Accessible *acc)
+{
+    AccessibleRelation **relation;
+    gint i;
+    gboolean rv = FALSE;
+
+    srl_assert (acc && label);
+
+    relation = Accessible_getRelationSet (acc);
+    if (!relation)
+	return FALSE;
+    for (i = 0; !rv && relation[i]; ++i)
+    {
+	if (AccessibleRelation_getRelationType (relation[i]) == SPI_RELATION_LABELED_BY)
+	{
+	    gint j, cnt;
+	    cnt = AccessibleRelation_getNTargets (relation[i]);
+	    for (j = 0; !rv && j < cnt; j++)
+	    {
+		Accessible *label_;
+		label_ = AccessibleRelation_getTarget (relation[i], j);
+		if (label_ == label)
+		    rv = TRUE;
+		if (label_)
+		    Accessible_unref (label_);
+	    }
+	}
+    }
+    for (i = 0; relation[i]; ++i)
+	AccessibleRelation_unref (relation[i]);
+    g_free (relation);
+
+    return rv;
+}
+
+
+#define SRL_MAX_LEVELS_UP 5
+static Accessible*
+srl_get_context (Accessible *acc)
+{
+    Accessible *context;
+    gint i;
+
+    sru_assert (acc);
+
+    context = NULL;
+    Accessible_ref (acc);
+    for (i = 0; i < SRL_MAX_LEVELS_UP; i++)
+    {
+	Accessible *parent = Accessible_getParent (acc);
+	if (parent)
+	{
+	    AccessibleRole role;
+	    Accessible_unref (acc);
+	    acc = parent;
+	    role = Accessible_getRole (acc);
+	    if (role == SPI_ROLE_EMBEDDED || role == SPI_ROLE_TOOL_BAR)
+	    {
+		context = acc;
+		Accessible_ref (context);
+	    }
+	    if (srl_acc_has_state (acc, SPI_STATE_VERTICAL) && 
+		    (role == SPI_ROLE_FILLER || role == SPI_ROLE_PANEL))
+	    {
+		gint cnt;
+		cnt = Accessible_getChildCount (acc);
+		if (cnt == 2)
+		{
+		    Accessible *label, *container;
+		    label = Accessible_getChildAtIndex (acc, 0);
+		    container = Accessible_getChildAtIndex (acc, 1);
+		    if (label && container)
+		    {
+			if (Accessible_getRole (label) == SPI_ROLE_LABEL &&
+			    Accessible_getRole (container) == SPI_ROLE_FILLER)
+			{
+			    gchar *name = Accessible_getName (label);
+			    if (name && name[0])
+				if (!srl_is_label_for (label, acc))
+				{
+				    context = label;
+				    Accessible_ref (label);
+				}
+			    SPI_freeString (name);
+			}
+		    }
+		    if (label)
+		        Accessible_unref (label);
+		    if (container)
+		        Accessible_unref (container);
+		}
+	    }
+	}
+	else
+	    break;
+	if (context)
+	    break;
+    }
+    if (acc)
+	Accessible_unref (acc);
+    return context;
+}
+
+static gboolean
+srl_check_context_changed (SRLEvent *event)
+{
+    Accessible *context;
+
+    srl_assert (event);
+
+    context = srl_get_context (event->acc_ev->source);
+    if (context)
+    {
+	if (context != srl_last_context)
+	{
+	    SRLEvent *ev;
+	    if (srl_last_context)
+		Accessible_unref (srl_last_context);
+	    srl_last_context = context;
+	    ev = srle_new ();
+	    ev->acc = srl_last_context;
+	    Accessible_ref (srl_last_context);
+	    ev->type = SRL_EVENT_CONTEXT_SWITCHED;
+	    srl_notify_clients_obj (ev, SR_EVENT_WINDOW);
+	    srle_free (ev);
+	}
+	else
+	{
+	    Accessible_unref (context);
+	}
+    }
+    else
+    {
+	if (srl_last_context)
+	    Accessible_unref (srl_last_context);
+	srl_last_context = NULL;    
+    }
+    return TRUE;
+}
+
+
 static gboolean
 srl_process_event (SRLEvent *event)
 {
@@ -1534,6 +1890,9 @@ srl_process_event (SRLEvent *event)
     srle_set_acc (event);
     srle_change_type (event);
 
+    if (srle_has_type (event, SRL_EVENT_FOCUS))
+	srl_check_context_changed (event);
+
     if (srle_has_type (event, SRL_EVENT_FOCUS) ||
 	srle_has_type (event, SRL_EVENT_FOCUS2))
 	    srl_set_last_focus ((Accessible*)srle_get_acc (event));
@@ -1543,7 +1902,13 @@ srl_process_event (SRLEvent *event)
 /* FIXME: 
     if (srle_has_type (event, SRL_EVENT_WINDOW_DEACTIVATE))
 	report no focus if no other event */
-    process = srl_event_is_reported (event);
+    process = srl_event_is_reported (event);	
+    
+    if (srle_has_type (event, SRL_EVENT_FOCUS) ||
+	srle_has_type (event, SRL_EVENT_FOCUS2) ||
+	srle_has_type (event, SRL_EVENT_FOCUS1))
+	    srl_set_last_focus2 ((Accessible*)srle_get_acc (event));
+	
     if (process)
 	srl_report_event (event);
 
@@ -1613,11 +1978,14 @@ srl_init ()
     for (i = 0 ; i < SRL_EVENT_PRIORITY_LAST; i++)
 	srl_last_events[i] = NULL;
     srl_last_focus	= NULL;
+    srl_last_focus2    	= NULL;
     srl_last_edit	= NULL;
     srl_idle_need	= FALSE;
     srl_idle_installed	= FALSE;
     srl_watched_acc	= NULL;
     srl_last_table	= NULL;
+    srl_last_context 	= NULL;
+    srl_last_create 	= NULL;
     
     srl_log_init ();
     srl_timing_init ();
@@ -1631,18 +1999,20 @@ srl_init ()
 	    rv = SPI_registerGlobalEventListener (srl_event_listeners[i],
 				srl_events_type_name[i].name);
 	    if (!rv)
-	    	srl_error ("Cannot register a listener for event \"%s\".",
+	    {
+		gdk_beep ();
+	    	srl_warning ("Cannot register a listener for event \"%s\".",
 				srl_events_type_name[i].name);
+	    }
 	}
 	else
 	{
-	    srl_error ("Cannot create a listener for event \"%s\"",
+	    srl_warning ("Cannot create a listener for event \"%s\"",
 				srl_events_type_name[i].name);		
-	    rv = FALSE;
+	    rv = TRUE;
 	}
-	if (!rv)
-	    break;
     }
+    rv = TRUE;
 
     if (rv)
 	srl_set_initialized ();
@@ -1672,12 +2042,19 @@ srl_terminate ()
 	    srle_free (srl_last_events[i]);
     if (srl_last_focus)
 	Accessible_unref (srl_last_focus);
+    if (srl_last_focus2)
+	Accessible_unref (srl_last_focus2);	
     if (srl_last_edit)
         Accessible_unref (srl_last_edit);
     if (srl_watched_acc)
 	Accessible_unref (srl_watched_acc);
     if (srl_last_table)
 	Accessible_unref (srl_last_table);
+    if (srl_last_context)
+	Accessible_unref (srl_last_context);
+    if (srl_last_create)
+	Accessible_unref (srl_last_create);
+
     srl_log_terminate ();
     srl_timing_terminate ();
     srl_set_uninitialized ();
@@ -1807,3 +2184,19 @@ srl_unwatch_all_objects ()
 	Accessible_unref (srl_watched_acc);
     srl_watched_acc = NULL;
 }
+
+gboolean
+srl_is_object_watched (SRObject *obj)
+{
+    srl_assert (obj);
+
+    return sro_get_acc (obj) == srl_watched_acc;
+}
+
+gboolean
+srl_is_object_focused (SRObject *obj)
+{
+    srl_assert (obj);
+
+    return sro_get_acc (obj) == srl_last_focus || sro_get_acc (obj) == srl_last_focus2;
+}
